/* Programm zur Lsung der Aufgabe 4.8c

	1. Windows: Kompilieren mit dem freien Kompiler Dev-C++
		- Dev-C++ installieren 
		- Lsg-4-8c.dev Projekt-File im Windows-Explorer ffnen
		- F9 (Kompilieren und Starten)
	2. Linux/Solaris: Kompilieren mit dem freien gcc 
		- g++ -o lsg-4-8c main.cpp 
		- File lsg-4-8c ausfhren

	Autor: Ingo Boersch, April 2007, V1.0  */

#include <iostream>
using namespace std;

#define POP_SIZE	  50               // Populationsgroesse (muss durch 2 teilbar sein, da Nachkommen paarweise erzeugt werden)
#define MAXGEN	      4000000          // Maximale Anzahl von Generationen
#define MUTATIONSRATE_VON 1            // untere Grenze d. Mutationswahrscheinlichkeit auf Zeichenebene in Promille
#define MUTATIONSRATE_BIS 20           // untere Grenze d. Mutationswahrscheinlichkeit auf Zeichenebene in Promille
#define EXPERIMENTS   5                // Anzahl von Versuchen pro Mutationswahrscheinlichkeit
#define DEBUG         0                // Werte ungleich 0 fhren zur Ausgabe von Fitnesshistogrammen der Generationen

char ZielString[] = "Das ist ein guter Tag";
long int results[MUTATIONSRATE_BIS-MUTATIONSRATE_VON+1][EXPERIMENTS];
#define ALLELS      21                 // Lnge der Zielzeichenkette

struct Individuum {
	unsigned char Allele[ALLELS];      // Genotyp - die Zeichenkette
	int Fitness;                       // Fitnesswert, Anzahl korrekter Zeichen
};

class CPopulation {
public:
	// ffentliche Methoden   
	int  Starte_Evolution();						
	void Ausgabe_Gewinner();
	int  GetGenerations();
	void SetMutationsrate(unsigned int);

protected:
	// Interne Methoden
	void Init();
	void Ausgabe_Individuum(int);
	void BerechneFitness(Individuum *);          
	void Evaluate();
	void ErzeugeNaechsteGeneration();
	void ErzeugeNachkommen(int p1, int p2, Individuum *, Individuum *);
	int  Roulette();
	char Zufallszeichen();

	// Debugfunktionen
	void ShowHisto();

	// Variablen
	Individuum population[POP_SIZE];					// Population, Array der Individuen
	long int generations;								// Anzahl evolvierter Generationen
	int BestSoFarFitness;                               // bester ermittelter Fitnesswert 
	int BestSoFarIndex;                                 // Index des Individuums mit bester Fitness
	float FitnessSumme;                                 // Summe der Fitness der aktuellen Generation   
	unsigned int Mutationsrate;                         // Mutationsrate auf Zeichenebene in Promille
};

// Hauptmethode der Population: liefert 1, wenn ein optimales Individuum innerhalb von MAXGEN Generationen
// erzeugt werden konnte, andernfalls 0
int CPopulation::Starte_Evolution() {
	Init();                                     // Erzeuge zufllige, initiale Generation
	Evaluate();                                 // Evaluation der ersten Generation
	if (BestSoFarFitness == ALLELS) return 1;   // Schon eine optimale Lsung mit maximaler Fitness dabei?

	// Schleife bis zur optimalen Lsung, aber hchstens MAXGEN-mal
	while (BestSoFarFitness != ALLELS && generations < MAXGEN) {	
		ErzeugeNaechsteGeneration();
		Evaluate();
		if (DEBUG) if (generations%10 == 0) ShowHisto();                // zur Visualisierung der Evolution, Debug
		generations++;
	}

	if (BestSoFarFitness == ALLELS) return 1;       // wurde optimale Lsung innerhalb von MAXGEN Generationen gefunden?
	else return 0;
}

// Ausgabe eines eventuellen Gewinnerindividuums
void CPopulation::Ausgabe_Gewinner() {
	if (BestSoFarIndex >= 0 && BestSoFarIndex < POP_SIZE) Ausgabe_Individuum(BestSoFarIndex);                        
}

// Liefert die Anzahl abgelaufener Generationen
int CPopulation::GetGenerations() {return generations;}

// Setzt die Mutationsrate
void CPopulation::SetMutationsrate(unsigned int m) {
     Mutationsrate = m;
}

// ------------- private Methoden -------------------
// Erzeuge initiale Population
void CPopulation::Init() {
	generations = 0;
	BestSoFarFitness = -1;                                             
	BestSoFarIndex = -1;        

	for(int i=0;i<POP_SIZE;i++) 	
		for (int j=0;j<ALLELS;j++) 
			population[i].Allele[j] = Zufallszeichen();
}     

// Ausgabe eines Individuums
void CPopulation::Ausgabe_Individuum(int k) {
	for(int i = 0; i < ALLELS; i++) cout << population[k].Allele[i] ;
	cout << endl;
}

/* Berechnet die Fitness (Anzahl der bereinstimmungen mit dem ZielString) eines Genes */
void CPopulation::BerechneFitness(Individuum *indi) {
	indi->Fitness = 0;
	for(int i = 0; i < ALLELS;i++) if (indi->Allele[i] == ZielString[i]) indi->Fitness++;
}

// Berechne die Fitness aller Individuen und erzeuge eine Ausgabe bei neuer Bestleistung
void CPopulation::Evaluate() {
	FitnessSumme = 0;
	for(int i=0; i<POP_SIZE; i++) {
		BerechneFitness(&population[i]);

		// Neuen Spitzenreiter gefunden?
		if (population[i].Fitness > BestSoFarFitness) {
			BestSoFarFitness = population[i].Fitness;
			BestSoFarIndex = i;
			cout << "Gen "<< generations << " F=" << population[i].Fitness << ":"; Ausgabe_Individuum(i);
		}
		FitnessSumme += population[i].Fitness;   // Summe aller Fitnesswerte fr das sptere Roulette berechnen
	}
}

// Erzeugt eine neue Generation und ersetzt die alte
void CPopulation::ErzeugeNaechsteGeneration() {
	Individuum temppop[POP_SIZE];
	int parent1, parent2; 
	int i;

	for( i=0;i<POP_SIZE;i+=2) {
		parent1 = Roulette();
		parent2 = Roulette();
		ErzeugeNachkommen(parent1, parent2,&temppop[i], &temppop[i+1]);		
	}

	// Ersetze alte Generation
	for(i=0;i<POP_SIZE;i++) population[i] = temppop[i];
}

// Erzeugt aus 2 Eltern 2 Nachkommen mittels Crossover und Mutation
void CPopulation::ErzeugeNachkommen(int p1, int p2, Individuum *child1, Individuum *child2) {
	int i;
	int crossover_point = rand() % (ALLELS+1);    // Crossover-Punkt bestimmen

	// Crossover
	for(i=0;i<crossover_point;i++)      { 
		child1->Allele[i] = population[p1].Allele[i];
		child2->Allele[i] = population[p2].Allele[i];
	}
	for(i=crossover_point;i<ALLELS;i++) { 
		child1->Allele[i] = population[p2].Allele[i];
		child2->Allele[i] = population[p1].Allele[i];
	}

	// Mutation 
	for(i=0;i<ALLELS;i++) {	if (rand() % 1001 < Mutationsrate) child1->Allele[i] = Zufallszeichen();}
	for(i=0;i<ALLELS;i++) {	if (rand() % 1001 < Mutationsrate) child2->Allele[i] = Zufallszeichen();}
}

// Whlt zufllig ein Individuen entsprechend relativer Fitness. Hierbei stellen wir uns die Fitnesswerte der einzelnen
// Individuen als Intervalle entsprechender Lnge hintereinandergelegt vor, auf die eine Zufallszahl geworfen wird.
int CPopulation::Roulette() {
	float Obere_Intervallgrenze;
	float Zufallszahl;
	int i;

	Zufallszahl = rand() / (float)RAND_MAX * FitnessSumme;     // Zufallszahl von 0 .. FitnessSumme
	Obere_Intervallgrenze = population[0].Fitness;
	for(i = 0; i < POP_SIZE-1; i++) {
		// Fllt die Zufallszahl in das Intervall dieses Individuums i?
		if (Zufallszahl <= Obere_Intervallgrenze) return i;
		else Obere_Intervallgrenze+= population[i+1].Fitness;
	}
	return POP_SIZE-1; 
}

// Erzeuge zuflliges ASCII-Zeichen aus dem Zeichenvorrat "a..z,A..Z,Space"
char CPopulation::Zufallszeichen() {
	unsigned char i;
	i = rand() % 53;
	if (i == 0) return ' ';
	else if (i < 27) return 'a'-1+i;
	else return 'A'-27+i;
}

// Gibt ein Histogramm der Fitnesswerte einer Generation aus
void CPopulation::ShowHisto() {
	int histo[ALLELS+1];

	for( int i = 0; i< ALLELS+1;i++) {histo[i]=0;}
	for(int i=0;i<POP_SIZE;i++) histo[population[i].Fitness]++;

	cout << "Gen " << generations <<":";
	for( int i = 0; i< ALLELS+1;i++) {
		if (histo[i] == 0) cout << ". ";
		else cout << histo[i] << " ";
	}
	cout << endl;
}

// Hauptprogramm
int main() {
	int result;
	unsigned int m,v;
	CPopulation pop;
	
	srand(time(NULL));                                    // Init des Zufallszahlengenerators

	for( m = MUTATIONSRATE_VON; m <= MUTATIONSRATE_BIS ; m++) {
      pop.SetMutationsrate(m);
      for ( v=0;v<EXPERIMENTS;v++) {
          if (pop.Starte_Evolution()) {
		     cout << endl;
		     pop.Ausgabe_Gewinner();
		     cout << "Anzahl Generationen: " << pop.GetGenerations() << endl;
		     results[m-MUTATIONSRATE_VON][v]=pop.GetGenerations();
           }
	      else 
		     cout << endl << "Keine Loesung mit "<< MAXGEN << " Generationen gefunden." << endl;
   	      results[m-MUTATIONSRATE_VON][v]=pop.GetGenerations();
       }
    }
	
    cout << endl << "Ergebnis des Experiments mit " << POP_SIZE << " Individuen und " << endl << "Pm von " << MUTATIONSRATE_VON << " bis " << MUTATIONSRATE_BIS << " Promille (1. Spalte) mit je " << EXPERIMENTS << " Versuchen:" << endl;
    for( m = MUTATIONSRATE_VON; m <= MUTATIONSRATE_BIS ; m++) {
      cout << endl << m << " ";
      for ( v=0;v<EXPERIMENTS;v++) {
          cout << results[m-MUTATIONSRATE_VON][v] << "  ";
       }
    }
	
    getchar();
	return 0;
}
